import asyncio
import time

import aiohttp
import requests
import urllib3

import pcoss_scheduler_pkg.constants as c
import pcoss_scheduler_pkg.output
from pcoss_scheduler_pkg.helpers import Measurement, scheduled_operation

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


async def operation(so: scheduled_operation, session, ts):
    it0 = time.perf_counter()
    me = Measurement(
        admission_time=it0,
        server_ret=None,
        request_processing_time=None,
        release_time=None,
        )

    url = f'{so.base_url}/{so.endpoint}/{so.machine}/{int(so.item)}/{int(so.operation_duration-1000)}'
    
    await asyncio.sleep(so.start_time/1000.0)

    it1 = time.perf_counter()
    async with session.get(url, ssl=False, timeout=600) as resp:
        it2 = time.perf_counter()
        ret = await resp.text()

    it3 = time.perf_counter()

    me.server_ret = float(ret)
    me.request_processing_time = (it2 - it1) * 1000
    me.release_time = it3

    if me.request_processing_time > so.operation_duration:
        _correct_schedule(so)

    if c.PRINT_RESPONSES:
        print(f'{url=}\t{ret=}')
    
    if c.PRINT_RESPONES_DEBUG_TIMES:
        error_symbol = '***'
        ok_symbol = ''
        print(f'{so.endpoint=}\t{so.machine=}\t{so.start_time=:.2f}\t{so.end_time=:.2f}\t{so.operation_duration=:.2f}')
        print(f'{error_symbol if me.request_processing_time > so.operation_duration else ok_symbol}\tit2-it1={me.request_processing_time:.2f}')
    
    output.output_dict.update({(so.job, so.machine): (so, me)})

    return ret


def _correct_schedule(so: scheduled_operation):
    if c.PRINT_SCHEDULE_CORRECTION_MESSAGE:
        print('Schedule needs correction')  
    else:
        pass



import random
from typing import Dict, List, Set, Tuple

import networkx as nx
import numpy as np
import pandas as pd

import pcoss_scheduler_pkg.constants as c
import pcoss_scheduler_pkg.helpers as h
from pcoss_scheduler_pkg.helpers import scheduled_operation as so
from pcoss_scheduler_pkg.problem_input import ProblemInput


class ScheduleAlgorithmBase:
    def __init__(self, processing_times: pd.DataFrame,
            conflict_graph: nx.Graph, problem_input: ProblemInput):

        self.problem_input = problem_input
        self.pt = processing_times
        self.conflict_graph = conflict_graph
        self.candidate_schedules = [np.full(self.pt.shape, np.nan)]
        self.row_conflicts_set = {
            (o0[1], o1[1])
            for (o0, o1) in filter(
                lambda e: e[0][0] == e[1][0],
                conflict_graph.edges())
                }

    def insertion_order(self) -> List[Tuple[int, int]]:
        raise NotImplementedError

    def run(self) -> List[so]:
        raise NotImplementedError

    outcome_graph: nx.DiGraph = None


class Randomized(ScheduleAlgorithmBase):
    def insertion_order(self):
        return np.random.permutation(list(np.ndindex(self.pt.shape)))

    def run(self):
        return [
            (random.random() * 5, (r_idx, c_idx))
            for (r_idx, r) in enumerate(self.pt)
            for (c_idx, c) in enumerate(self.pt)
            ]


class InsertionBeam(ScheduleAlgorithmBase):
    def insertion_order(self) -> List[Tuple[int, int]]:
        first_order = []
        min_shape = min(self.pt.shape)
        pt_trimmed = self.pt[:min_shape, :].copy()
        row_idx = 0

        for row_idx in range(min_shape):
            max_col_idx = np.argmax(pt_trimmed[row_idx, :])
            first_order += [(row_idx, max_col_idx)]
            pt_trimmed = np.delete(pt_trimmed, max_col_idx, axis=1)

        other_order = [(x[0], x[1]) for x in filter(
            lambda e: (e[0], e[1]) not in first_order,
            np.transpose(
                np.unravel_index(
                    np.argsort(self.pt, axis=None)
                    [::-1],
                    shape=self.pt.shape)))]

        return first_order + other_order

    def solve_conflicting_ranks(self, rm: np.ndarray,
            changed_rows: Set[int], changed_cols: Set[int]) -> np.ndarray:

        new_changed_rows = set()
        new_changed_cols = set()

        for cr in changed_rows:
            for cc in changed_cols:
                cv = rm[cr, cc]

                for n in filter(
                        lambda n: rm[n] == cv,
                        self.conflict_graph[(cr, cc)]):

                    new_changed_rows.add(n[0])
                    new_changed_cols.add(n[1])
                    rm[n] += 1

        if new_changed_rows:
            return self.solve_conflicting_ranks(
                rm,
                new_changed_rows,
                new_changed_cols)

        return rm

    def row_conflicts(self, rm: np.ndarray, row: int, col: int
            ) -> List[Tuple[int, int]]:

        ret = []

        for cp in self.row_conflicts_set:
            if col == cp[0] and not np.isnan(rm[row, cp[1]]):
                ret += [(row, cp[1])]
            if col == cp[1] and not np.isnan(rm[row, cp[0]]):
                ret += [(row, cp[0])]

        return ret

    def path_rec(self, rm: np.ndarray, e: Tuple[int, int], asc: bool
            ) -> List[Tuple[int, int]]:

        step = 1 if asc else -1
        next_val = rm[e] + step
        next_e = next(
            filter(lambda n: rm[n] == next_val, self.conflict_graph[e]),
            None)

        if next_e:
            return [next_e] + self.path_rec(rm, next_e, asc)

        return []

    def path_while(self, rm: np.ndarray, e: Tuple[int, int], asc: bool
            ) -> List[Tuple[int, int]]:

        cg = self.conflict_graph
        step = 1 if asc else -1
        ret = []
        next_e = e

        while next_e:
            next_val = rm[next_e]+step
            next_e = next(
                filter(lambda n: rm[n] == next_val, cg[next_e]),
                None)

            if next_e:
                ret += [next_e]

        return ret

    def get_path(self, rm: np.ndarray, added_element_idx: Tuple[int, int]
            ) -> List[Tuple[int, int]]:

        pr = self.path_rec

        return (
            pr(rm, added_element_idx, False)
            + [added_element_idx]
            + pr(rm, added_element_idx, True)
        )

    def beam_search(self, rm_list: List[np.ndarray],
            added_element_idx: Tuple[int, int]) -> List[np.ndarray]:

        rm_costs = {}

        for rm_idx, rm_cur in enumerate(rm_list):
            jobs, machines = np.transpose(
                self.get_path(rm_cur, added_element_idx))
            cost = sum(self.pt[jobs, machines])
            rm_costs[rm_idx] = cost

        return (
            [
                rm_list[k]
                for k, v in
                sorted(rm_costs.items(), key=lambda e: e[1])
            ]
            [:self.problem_input.beam_width]
            )

    def schedule_as_graph(self, rm: np.ndarray):
        max_rank = int(np.max(rm))

        G = nx.DiGraph()

        for from_rank in range(1, max_rank):
            to_rank = from_rank + 1
            to_idxs = np.nonzero(rm == to_rank)

            for to_idx in np.transpose(to_idxs):
                same_job_idxs = [
                    (to_idx[0], machine)
                    for machine in np.nonzero(rm[to_idx[0], :] == from_rank)[0]
                ]
                same_machine_idxs = [
                    (job, to_idx[1])
                    for job in np.nonzero(rm[:, to_idx[1]] == from_rank)[0]
                ]

                for from_idx in same_job_idxs + same_machine_idxs:
                    G.add_edge(from_idx, tuple(to_idx))

        self.outcome_graph = G

        if self.problem_input.show_result_schedule_graph:
            h.plot_schedule_graph(self.outcome_graph)

    def schedule_as_list_of_scheduled_operations(self, rm: np.ndarray
            ) -> List[so]:

        scheduled_operations: List[so] = []
        cur_rank = 1
        st_mx = np.full(rm.shape, np.nan, dtype=object)

        while len(scheduled_operations) < rm.size:
            idxs = np.nonzero(rm == cur_rank)

            for idx in np.transpose(idxs):
                if cur_rank == 1:
                    prev_max_time = 0
                else:
                    prev_max_time = max(
                        [s.end_time
                        for s in st_mx[
                            tuple(np.transpose([e[0]
                            for e in self.outcome_graph.in_edges([tuple(idx)])]))]])

                end_time = prev_max_time + self.pt[idx[0]][idx[1]]

                cur_so = so(
                    base_url=self.problem_input.base_url,
                    rank=cur_rank,
                    start_time=prev_max_time,
                    end_time=end_time,
                    operation_duration=self.pt[idx[0]][idx[1]],
                    job=idx[0],
                    machine=idx[1],
                    endpoint=None,
                    item=None,
                    )
                scheduled_operations.append(cur_so)
                st_mx[idx[0]][idx[1]] = cur_so

            cur_rank += 1

        return scheduled_operations

    def run(self):
        for job, machine in self.insertion_order():
            candidate_schedules_with_children = []
            for cs in self.candidate_schedules:
                potential_rank = set()

                potential_rank.add(1)
                for e in np.argwhere(~np.isnan(cs[:, machine])):
                    potential_rank.add(cs[(e, machine)][0] + 1)
                for cell in self.row_conflicts(cs, job, machine):
                    potential_rank.add(cs[cell] + 1)

                for pr in potential_rank:
                    cs_with_child = cs.copy()
                    cs_with_child[job, machine] = pr
                    cs_with_child = self.solve_conflicting_ranks(
                        cs_with_child, {job,}, {machine,})
                    candidate_schedules_with_children.append(cs_with_child)

            self.candidate_schedules = self.beam_search(
                candidate_schedules_with_children, (job, machine))

        self.schedule_as_graph(self.candidate_schedules[0])

        return self.schedule_as_list_of_scheduled_operations(
            self.candidate_schedules[0])


ALGORITHM_CLASS_DICT = {
    'randomized': Randomized,
    'insertion_beam': InsertionBeam,
}



import asyncio
# import sched
import time
from dataclasses import asdict
from typing import Dict, List, Tuple

import aiohttp
import networkx as nx
import numpy as np
import pandas as pd

import pcoss_scheduler_pkg.algorithms as a
import pcoss_scheduler_pkg.constants as c
import pcoss_scheduler_pkg.helpers as h
import pcoss_scheduler_pkg.operations as o
from pcoss_scheduler_pkg.problem_input import ProblemInput


class Scheduler:
    def __init__(self, table: pd.DataFrame):
        self.problem_input = None

        self.table: pd.DataFrame = table
        self._table: pd.DataFrame = table.to_numpy()
        self.row_grouping_func = h.id_func
        self.cloumn_grouping_func = h.id_func
        self.column_operation_dict: Dict = {}
        self.conflict_graph: nx.Graph = None
        self._conflict_graph: nx.Graph = None
        self.complexity_table: pd.DataFrame = None
        self.processing_times: pd.DataFrame = None
        self._processing_times: pd.DataFrame = None
        
        self.algorithm = None
        self.objective = c.DEFAULT_OBJECTIVE
        self.beam_witdh = c.DEFAULT_BEAM_WIDTH

        self.schedule: List[h.scheduled_operation] = []
        
        self._session = aiohttp.ClientSession()

    def _find_best_algorithm(self):
        if self.algorithm:
            return
        
        self.algorithm = c.DEFAULT_ALGORITHM

    def _group_rows(self):
        self._table = self.row_grouping_func(self._table)

    def _group_columns(self):
        self._table = self.cloumn_grouping_func(self._table)

    def _fill_times(self):
        if not self.processing_times.empty:
            self._processing_times = self.processing_times.to_numpy()
            return

        if self.complexity_table:
            self._processing_times = self._table * self.complexity_table
            return
        
        self._processing_times = self._assses_aproximate_execution_times()

    def _assses_aproximate_execution_times(self):
        # todo
        return np.ones(self._table.shape)

    @h.get_func_exec_time_decorator
    def _prepare_data(self):
        self._group_rows()
        self._group_columns()
        self._fill_times()
        self._find_best_algorithm()
    
    @h.get_func_exec_time_decorator
    def _prepare_schedule(self):
        self.schedule = (
            a.ALGORITHM_CLASS_DICT[self.algorithm]
            (self._processing_times, self.conflict_graph, self.problem_input)
            .run()
        )

    def prepare(self):
        self._prepare_data()
        self._prepare_schedule()

    async def run(self):
        t0 = time.perf_counter()

        al = []
        for so in self.schedule:
            so.endpoint = '0' 
            so.item = self._table[so.job][so.machine]
            
            al.append(o.operation(so, self._session, t0))
            
        t1 = time.perf_counter()
        await asyncio.gather(*al)
        t2 = time.perf_counter()
        try:
            await self._session.close()
        except Exception as e:
            pass
        t3 = time.perf_counter()
        if c.PRINT_DEBUG_MESSSAGES:
            print(f't1-t0: {t1-t0}\tt2-t1: {t2-t1}\tt3-t2: {t3-t2}\tt3-t0: {t3-t0}')

    @classmethod
    def from_ProblemInput(cls, problem_input: ProblemInput):
        s = cls(problem_input.table_in)

        s.problem_input = problem_input
        if problem_input.column_operation_dict:
            s.column_operation_dict = problem_input.column_operation_dict
        if problem_input.conflict_graph_in:
            s.conflict_graph = problem_input.conflict_graph_in
        if not problem_input.processing_times.empty:
            s.processing_times = problem_input.processing_times
        if problem_input.algorithm:
            s.algorithm = problem_input.algorithm
        if problem_input.objective:
            s.objective = problem_input.objective
        if problem_input.beam_width:
            s.beam_width = problem_input.beam_width

        return s



import argparse
import asyncio
import time
from typing import List, Set, Tuple

import pcoss_scheduler_pkg.constants as c
import pcoss_scheduler_pkg.helpers as h
import pcoss_scheduler_pkg.output
import pcoss_scheduler_pkg.scheduler as s
from pcoss_scheduler_pkg.problem_input import ProblemInput


async def main():
    ap = argparse.ArgumentParser()
    ap.add_argument("toml_file", help="Path to the TOML file to be used", type=str)
    args = ap.parse_args()

    pi = ProblemInput.from_toml(args.toml_file)
    c.PRINT_METHOD_TIMES = pi.print_method_times
    
    sc = s.Scheduler.from_ProblemInput(pi)

    sc.prepare()
    
    if pi.show_gantt_plot:
        h.plot_gantt_chart(sc.schedule)

    ts = time.perf_counter()
    await sc.run()
    te = time.perf_counter()

    if pi.print_method_times:
        print(f'Function "run" execution time: {te-ts:.3f}s')

    # print(output.output_dict)


if __name__ == '__main__':
    (asyncio
        .get_event_loop()
        .run_until_complete(
            main()
        )
    )



from typing import Dict, List, Set, Tuple

import pandas as pd
import toml

import pcoss_scheduler_pkg.constants as c
import pcoss_scheduler_pkg.helpers as h


class ProblemInput:
    def __init__(self, init_dict: Dict):

        self.table_in = pd.read_csv(
            init_dict['files']['data'],
            index_col=init_dict['problem_data']['index_cols'],
            usecols=lambda col: col not in init_dict['problem_data']['grouping_cols'],
            )

        self.column_operation_dict = {}
        avaliable_operations_list = ['A', 'B', 'C']

        use_operation_num = 0

        for cm in init_dict['problem_data']['conflicting_machines']:
            self.column_operation_dict[cm[0]] = avaliable_operations_list[use_operation_num]
            self.column_operation_dict[cm[1]] = avaliable_operations_list[use_operation_num]
            use_operation_num = (use_operation_num + 1) % 3

        for col_num in range(len(self.table_in.columns)):
            if col_num not in [cm for te in init_dict['problem_data']['conflicting_machines'] for cm in te]:
                self.column_operation_dict[col_num] = '0'

        job_cnt, machine_cnt = self.table_in.shape

        self.conflict_graph_in = (
            h.create_conflict_graph_from_cnts_and_conflicting_machine_list(
                job_cnt=job_cnt,
                machine_cnt=machine_cnt,
                conflicting_machines=init_dict['problem_data']['conflicting_machines'],
                show=init_dict['display_config']['show_conflict_graph']))

        self.processing_times = pd.read_csv(
            init_dict['files']['processing_times'],
            index_col=init_dict['problem_data']['index_cols'],
            usecols=lambda col: col not in init_dict['problem_data']['grouping_cols'],
            ) * 100 + 1000
        
        self.conflicting_machines_list = init_dict['problem_data']['conflicting_machines']

        self.base_url = init_dict['problem_data']['base_url']

        self.algorithm = init_dict['algorithm_config']['algorithm_name']
        self.objective = init_dict['algorithm_config']['objective']
        self.beam_width = init_dict['algorithm_config']['beam_width']

        self.print_responses = init_dict['display_config']['print_responses']
        self.print_method_times = init_dict['display_config']['print_method_times']
        self.show_conflict_graph = init_dict['display_config']['show_conflict_graph']
        self.show_result_schedule_graph = init_dict['display_config']['show_result_schedule_graph']
        self.show_gantt_plot = init_dict['display_config']['show_gantt_plot']


    @classmethod
    def from_toml(cls, toml_file: str):
        config_dict = toml.load(toml_file)
        
        config_dict['problem_data']['conflicting_machines'] = list(
            map(tuple, config_dict['problem_data']['conflicting_machines'])
            )
        
        return cls(config_dict)



import time
from dataclasses import asdict, dataclass
from functools import wraps
from random import random
from typing import List, Set, Tuple

import networkx as nx
from matplotlib import pyplot as plt

import pcoss_scheduler_pkg.constants as c


@dataclass
class Measurement:
    admission_time: int
    server_ret: float
    request_processing_time: float
    release_time: int


@dataclass
class scheduled_operation:
    rank: int
    start_time: float
    end_time: float
    operation_duration: float
    job: int
    machine: int
    endpoint: str
    item: object
    base_url: str


def id_func(x):
    return x


def create_conflict_graph_from_cnts_and_conflicting_machine_list(
        job_cnt: int,
        machine_cnt: int,
        conflicting_machines: Set[Tuple[int, int]],
        show=True) -> nx.Graph:

    G = nx.Graph()
    for job_idx in range(job_cnt):
        job_nodes = [
            (job_idx, machine_idx)
            for machine_idx
            in range(machine_cnt)
            ]
        
        G.add_nodes_from(job_nodes)

        for cm1, cm2 in conflicting_machines:
            G.add_edge((job_idx, cm1), (job_idx, cm2))
        
        for prev_job_idx in range(job_idx):
            for machine_idx in range(machine_cnt):
                G.add_edge((prev_job_idx, machine_idx), (job_idx, machine_idx))

    if show:
        plt.figure(figsize=(6, 6))
        pos = {
            (x, y): (y + random() / 3, -x + random() / 3)
            for x, y in G.nodes()
            }
        
        nx.draw(G, with_labels=True, pos=pos, connectionstyle='arc3, rad=2')
        plt.show()
    
    return G


def get_func_exec_time_decorator(f: callable):
    @wraps(f)
    def wf(*args, **kwargs):
        st = time.perf_counter()
        f(*args, **kwargs)
        et = time.perf_counter()
        if c.PRINT_METHOD_TIMES:
            print(f'Function "{f.__name__}" exection time: {et-st:.3f}s')

    return wf


def plot_gantt_chart(schedule: List[scheduled_operation]):
    import pandas as pd
    import plotly.express as px

    def tramsform_opearation_time_to_date(operation_time: float):
        beginning_date = pd.to_datetime('2020-01-01')

        return beginning_date + pd.DateOffset(int(operation_time/1000))

    schedule_df = pd.DataFrame([asdict(so) for so in schedule])

    if c.PRINT_DEBUG_MESSSAGES:
        print(f'max et: {max(schedule_df.end_time)}')

    schedule_df.start_time = schedule_df.start_time.apply(
        tramsform_opearation_time_to_date)
    schedule_df.end_time = schedule_df.end_time.apply(
        tramsform_opearation_time_to_date)

    fig = px.timeline(
        schedule_df,
        x_start="start_time",
        x_end="end_time",
        y="machine",
        color="job",
        color_continuous_scale=px.colors.sequential.Greys)
        
    fig.show()


def plot_schedule_graph(graph: nx.DiGraph):
    plt.figure(figsize=(12, 12))
    pos = {
        (x, y): (y + random() / 3, -x + random() / 3)
        for x, y in graph.nodes()
        }
    
    nx.draw(graph, pos=pos, with_labels=True)
    plt.show()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()


with open('requirements.txt') as f:
    required = f.read().splitlines()

setuptools.setup(
    name="pcoss-scheduler-pkg-ksazon",
    version="0.0.18",
    author="Krzysztof Sazon",
    author_email="krzysztof.sazon@gmail.com",
    description="Tabular data PCOSS calculations module",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ksazon/mgr/mgr_client",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',
    install_requires=required
)



BASE_URL = r'https://localhost:44320/Operations'
DEFAULT_ALGORITHM = 'insertion_beam'
DEFAULT_OBJECTIVE = 'cmax'
DEFAULT_BEAM_WIDTH = 5

PRINT_RESPONSES = True
PRINT_RESPONES_DEBUG_TIMES = False
PRINT_METHOD_TIMES = False
PRINT_SCHEDULE_CORRECTION_MESSAGE = False
PRINT_DEBUG_MESSSAGES = False

SHOW_CONFLICT_GRAPH = False
SHOW_RESULT_SCHEDULE_GRAPH = False
SHOW_GANTT = False
